import os
import sys
import threading

from pytubefix.helpers import safe_filename
from pytubefix import Playlist

from pyutube.utils import console, asking_video_or_audio, ask_playlist_video_names, ask_for_make_playlist_in_order


class PlaylistHandler:
    playlist_videos = []

    def __init__(self, url: str, path: str):
        self.url: str = url
        self.path: str = path

    def process_playlist(self):
        """
        Process the playlist by asking for the audio or video, downloading the playlist,
        then asking for which video to download and downloading it.
        """
        console.print("Processing playlist...")

        try:
            is_audio = asking_video_or_audio()
        except TypeError:
            # If the user cancelled, return
            console.print("Cancelled")
            return

        console.print("Downloading playlist...")
        playlist = Playlist(self.url)
        print("playlist_stream", playlist.title)

        p_title = playlist.title
        p_total = playlist.length
        p_videos = playlist.videos

        make_in_order = ask_for_make_playlist_in_order()
        console.print("Fetching playlist videos...")
        self.get_all_playlist_videos_title(p_videos)

        if make_in_order:
            for index, video_and_id in enumerate(self.playlist_videos):
                new_video_title = f"{index + 1}__{video_and_id[0]}"
                self.playlist_videos[index] = (new_video_title, video_and_id[1])

        console.print("Checking if the videos are already downloaded...")
        new_path = self.check_for_downloaded_videos(p_title, p_total)

        console.print("Chose what video you want to download", style="info")
        videos_selected = ask_playlist_video_names(self.playlist_videos)

        return new_path, is_audio, videos_selected, make_in_order

    def fetch_title_thread(self, video, index, results):
        """
        Fetch all playlist video titles concurrently but maintain the order.
        """
        video_title = safe_filename(video.title)
        video_id = video.video_id
        results[index] = (video_title, video_id)

    def get_all_playlist_videos_title(self, videos):
        """
        Fetch all playlist video titles concurrently but maintain the order.
        """
        total_videos = len(videos)
        results = [None] * total_videos
        title_threads = []

        for index, video in enumerate(videos):
            thread = threading.Thread(
                target=self.fetch_title_thread, args=(video, index, results))
            thread.start()
            title_threads.append(thread)

        for thread in title_threads:
            thread.join()

        self.playlist_videos = results

    @staticmethod
    def show_playlist_info(title, total):
        console.print(f"\nPlaylist title: {title}\n", style="info")
        console.print(f"Total videos: {total}\n", style="info")

    def create_playlist_folder(self, title):
        os.makedirs(title, exist_ok=True)
        return os.path.join(self.path, title)

    def check_for_downloaded_videos(self, title, total):
        new_path = self.create_playlist_folder(safe_filename(title))

        # check if there is any video already downloaded in the past
        for file in os.listdir(new_path):
            for video in self.playlist_videos:
                if file.startswith(video[0]):
                    self.playlist_videos.remove(video)
                    # Exit the inner loop since we found a match
                    break

        if not self.playlist_videos:
            console.print(f"All playlist are already downloaded in this directory, see '{title}' folder", style="info")
            sys.exit()

        self.show_playlist_info(title, total)
        return new_path



"""
Pyutube is a command-line interface, a versatile tool
to download YouTube videos, shorts, and playlists.

This module provides a command-line interface (CLI), a powerful tool designed
to simplify the process of downloading YouTube content directly from the terminal.
Pyutube supports downloading videos (as video or audio), shorts, and playlists,
offering users flexibility and convenience in managing their media downloads.

Usage:
    $ pyutube download <URL> [options]

Options:
    -a, --audio          Download only audio.
    -f, --footage        Download only video (footage).
    -v, --version        Show the version number.

Example:
    $ pyutube <YouTube_URL> -a
        Download the audio of the specified YouTube video.

    $ pyutube <YouTube_URL> -f
        Download the video (footage) of the specified YouTube video.

    $ pyutube <YouTube_URL>
        Download the file of the specified YouTube video,
        it will ask you about downloading it as video or audio.

    $ pyutube <YouTube_playlist_URL>
        Download all videos from the specified YouTube playlist.

    $ pyutube <YouTube_short_URL>
        Download the specified YouTube short video.

Made with ❤️ By Ebraheem. Find me on GitHub: @Hetari. The project lives on @Hetari/pyutube.

Thank you for using Pyutube! Your support is greatly appreciated. ⭐️
"""

import os
import sys

import typer

from pyutube.utils import (
    __version__,
    __app__,
    clear,
    error_console,
    console,
    check_internet_connection,
    check_for_updates,
)
from pyutube.services import DownloadService
from pyutube.handlers import URLHandler

# Create CLI app
app = typer.Typer(
    name="pyutube",
    add_completion=False,
    help="Awesome CLI to download YouTube videos \
    (as video or audio)/shorts/playlists from the terminal",
    rich_markup_mode="rich",
)


# Define the variables for the arguments and options
url_arg = typer.Argument(
    None,
    help="YouTube URL [red]required[/red]",
    show_default=False
)
path_arg = typer.Argument(
    os.getcwd(),
    help="Path to save video [cyan]default: <current directory>[/cyan]",
    show_default=False
)
audio_option = typer.Option(
    False, "-a", "--audio", help="Download only audio"
)
video_option = typer.Option(
    False, "-f", "--footage", help="Download only video"
)
version_option = typer.Option(
    False, "-v", "--version", help="Show the version number"
)


@app.command(
    name="download",
    help="""
Download a [red]YouTube[/red] [green]videos[/green] [blue](as
video or audio)[/blue], [green]shorts[/green], and [green]playlists[/green].
    """,
    epilog="""
Made with ❤️  By Ebraheem. Find me on GitHub: [link=https://github.com/Hetari]@Hetari[/link].

The project lives on [link=https://github.com/Hetari/pyutube]@Hetari/pyutube[/link].\n\nThank you for using Pyutube! and your support :star:
""",
)
def pyutube(
    url: str = url_arg,
    path: str = path_arg,
    audio: bool = audio_option,
    video: bool = video_option,
    version: bool = version_option
) -> None:
    """
    Downloads a YouTube video.

    Args:
        url (str): The URL of the YouTube video.
        path (str): The path to save the video. Defaults to the current working directory.

    """
    check_for_updates()

    if version:
        console.print(f"Pyutube {__version__}")
        check_for_updates()
        sys.exit()

    if url is None:
        error_console.print("❗ Missing argument 'URL'.")
        sys.exit()

    clear()

    if not check_internet_connection():
        sys.exit()

    url_handler = URLHandler(url)
    is_valid_link, link_type = url_handler.validate()

    if not is_valid_link:
        sys.exit()

    download_service = DownloadService(url, path, None)
    if audio:
        download_service.is_audio = True
        video, video_id,  _, video_audio, _ = download_service.download_preparing()
        download_service.download_audio(video, video_audio, video_id)

    elif video or link_type == "short":
        video, video_id,  streams, video_audio, quality = download_service.download_preparing()
        video_file = download_service.video_service.get_video_streams(quality, streams)
        download_service.download_video(video, video_id, video_file, video_audio)

    elif link_type == "video":
        download_service.asking_video_or_audio()

    elif link_type == "playlist":
        download_service = DownloadService(url, path, None)
        download_service.get_playlist_links()

    else:
        error_console.print("❗ Unsupported link type.")
        sys.exit()

    sys.exit()



from termcolor import colored
from yaspin import yaspin
from yaspin.spinners import Spinners
from pytubefix import YouTube


class AudioService:
    def __init__(self, url: str):
        self.url = url

    @staticmethod
    @yaspin(
        text=colored("Downloading the audio...", "green"),
        color="green",
        spinner=Spinners.dots13
    )
    def get_audio_streams(video: YouTube) -> YouTube:
        """
        Function to get audio streams from a video.

        Args:
            video: The video for which audio streams are to be obtained.

        Returns:
            The first audio stream found in the video.
        """
        return video.streams.filter(only_audio=True).order_by('mime_type').first()



import os
import pytest
from unittest.mock import patch

from pyutube.utils import is_youtube_link
from pyutube.utils import is_youtube_video


@pytest.mark.parametrize("link, expected_result, test_id", [
    # Happy path tests
    ("https://www.youtube.com/watch?v=dQw4w9WgXcQ", True, "happy_standard"),
    ("http://www.youtube.com/watch?v=dQw4w9WgXcQ", True, "happy_http"),
    ("https://youtube.com/watch?v=dQw4w9WgXcQ", True, "happy_no_www"),
    ("https://youtu.be/dQw4w9WgXcQ", True, "happy_short"),
    ("www.youtube.com/watch?v=dQw4w9WgXcQ", True, "happy_no_protocol"),
    ("youtube.com/watch?v=dQw4w9WgXcQ", True, "happy_no_protocol_www"),
    ("youtu.be/dQw4w9WgXcQ", True, "happy_short_no_protocol"),
    ("https://www.youtube.com/watch?v=dQw4w9WgXcQ&t=100s",
     True, "happy_starts_at_100s"),

    ("https://www.youtube.com/watch?v=dQw4w9WgXcQ", True, "video_id_standard"),
    ("https://youtu.be/dQw4w9WgXcQ", True, "video_id_short"),
    ("http://www.youtube.com/watch?v=dQw4w9WgXcQ", True, "video_id_http"),
    ("www.youtube.com/watch?v=dQw4w9WgXcQ", True, "video_id_no_protocol"),
    ("youtube.com/watch?v=dQw4w9WgXcQ", True, "video_id_no_subdomain"),
    ("https://www.youtube.com/embed/dQw4w9WgXcQ", True, "video_embed_standard"),
    ("https://www.youtube.com/v/dQw4w9WgXcQ", True, "video_v_standard"),

    # Edge cases
    ("https://www.youtube.com/watch?v=dQw4w9WgXcQ&feature=youtu.be",
     True, "edge_with_params"),
    ("", False, "edge_empty_string"),
    ("https://www.youtube.com/watch?v=", False, "edge_no_video_id"),
    ("https://www.youtu.be/", False, "edge_short_no_video_id"),
    ("https://www.youtube.com/watch?", False, "edge_no_query_string"),

    ("", False, "empty_string"),
    ("https://www.youtube.com/watch?v=", False, "video_id_empty"),
    # ("https://www.youtube.com/playlist?list=", False, "playlist_id_empty"),
    ("https://www.youtu.be/", False, "short_domain_no_id"),
    ("https://youtube.com/watch?v=dQw4w9WgXcQ&feature=share",
     True, "video_id_with_query"),
    ("https://www.youtube.com/watch?v=dQw4w9WgXcQ&list=PLA7no0L9zTk4Qp6LlDjFVCetW-sFw9r5I",
     True, "video_and_playlist"),

    # Error cases
    ("https://www.notyoutube.com/watch?v=dQw4w9WgXcQ", False, "error_wrong_domain"),
    ("https://www.youtube.com/watchv=dQw4w9WgXcQ",
     False, "error_missing_question_mark"),
    ("https://youtu.be/dQw4w9WgXcQextra", True, "error_extra_characters"),
    ("https://www.youtube.com/watch?time_continue=1&v=dQw4w9WgXcQ",
     False, "error_unexpected_params"),

    ("https://example.com/watch?v=dQw4w9WgXcQ", False, "non_youtube_domain"),
    ("https://www.youtube.com/watchv=dQw4w9WgXcQ", False, "typo_in_query"),
    ("https://youtu.be/dQw4w9WgXcQ?list=PLA7no0L9zTk4Qp6LlDjFVCetW-sFw9r5I",
     True, "short_link_with_playlist"),
    ("https://www.youtub.com/watch?v=dQw4w9WgXcQ", False, "typo_in_domain"),
    ("https://www.youtube.com/watch?feature=share&v=dQw4w9WgXcQ",
     True, "video_id_with_other_query_first"),

])
def test_is_youtube_link(link, expected_result, test_id):
    # Act
    result = is_youtube_video(link)

    # Assert
    assert result == expected_result, f"Test failed for test_id: {test_id}"


@pytest.mark.parametrize("link, expected_result", [
    ("https://www.youtube.com/watch?v=dQw4w9WgXcQ",
     (True, "video")),
    ("https://youtu.be/dQw4w9WgXcQ", (True, "video")),
    # ("https://www.youtube.com/playlist?list=PLynG8gQD-n8BMplEGyFBRVJFssMFo5x07",
    #  (True, "playlist")),
    ("https://www.youtube.com/shorts/dQw4w9WgXcQ",
     (True, "short")),


    ("", (False, "unknown")),
    ("https://www.youtube.com/", (False, "unknown")),
    ("https://www.youtube.com/watch?v=", (False, "unknown")),
    ("https://www.youtu.be/", (False, "unknown")),
    ("https://www.youtube.com/playlist?list=",
     (False, "unknown")),
    ("https://www.youtube.com/shorts/", (False, "unknown")),


    ("https://www.google.com", (False, "unknown")),
    ("https://www.youtube.com/watch?list=dQw4w9WgXcQ",
     (False, "unknown")),
    ("https://www.youtu.be/dQw4w9WgXcQ?list=PLynG8gQD-n8BMplEGyFBRVJFssMFo5x07",
     (True, "video")),
    ("https://www.youtube.com/shorts/dQw4w9WgXcQ?list=PLynG8gQD-n8BMplEGyFBRVJFssMFo5x07",
     (True, "short")),
])
def test_is_youtube_link_error_cases(link, expected_result):
    # Act
    result = is_youtube_link(link)

    # Assert
    assert result == expected_result



import os
import sys
from pytubefix import YouTube
from pytubefix.helpers import safe_filename
from termcolor import colored

from pyutube.utils import ask_rename_file, error_console, console


class FileService:
    def save_file(self, video: YouTube, filename: str, path: str) -> None:
        """
        Save the file to the specified path with the given filename.

        Args:
            video: The video to be saved.
            path: The path where the video will be saved.
            filename: The name of the file.

        Returns:
            None
        """
        video.download(output_path=path, filename=filename)

    def generate_filename(self, video, video_id, is_audio=False, filename: str = ""):
        """
        Generate a filename for the downloaded video.

        Returns:
            str: The generated filename.
        """
        file_type = 'audio' if is_audio else video.resolution

        extension = '' if is_audio else f'.{video.mime_type.split('/')[1]}'

        title = filename if filename != "" else video.default_filename.split('.')[0]
        title = safe_filename(title)

        return f"{title}_{file_type}{extension}"

    def handle_existing_file(
            self, video: YouTube, video_id: str, filename: str, path: str, is_audio: bool = False) -> None:
        """
        Handle the case where a file with the same name already exists.

        Returns:
            str: The user's choice.
        """
        # If file with the same name already exists in the path
        if not self.is_file_exists(path, filename):
            return filename

        choice = ask_rename_file(filename).lower()
        if choice.startswith('rename'):
            filename = safe_filename(
                self.prompt_new_filename(filename)
            )
            if not filename:
                error_console.print("Invalid filename")
                return False
            filename = self.generate_filename(video, video_id, is_audio, filename)

            return filename
        elif choice.startswith('cancel'):
            console.print("Download canceled", style="info")
            sys.exit()

        # else overwrite
        return filename

    def prompt_new_filename(self, filename):
        """
        Prompt the user for a new filename.

        Returns:
            str: The new filename.
        """
        text = colored(filename, 'yellow')
        return input(f"Rename {text} to: ")

    @staticmethod
    def is_file_exists(path: str, filename: str) -> bool:
        """
        Check if a file exists at the specified path and filename.

        Args:
            path: The path where the file is located.
            filename: The name of the file to check.

        Returns:
            bool: True if the file exists, False otherwise.
        """
        return os.path.isfile(os.path.join(path, filename))



import re
import sys
from pyutube.utils import console, error_console


class URLHandler:
    def __init__(self, url):
        self.url = url

    def validate(self):
        if self.__is_youtube_video_id(self.url):
            self.url = f"https://www.youtube.com/watch?v={self.url}"

        return self.__validate_link(self.url)

    def __validate_link(self, url: str) -> tuple[bool, str]:
        """
        Validates the given YouTube video URL.

        Args:
            url (str): The URL of the YouTube video.

        Returns:
            Tuple[bool, str]: A tuple containing a boolean indicating if the link is valid
            and a string indicating the type of the link (video or short).
        """
        is_valid_link, link_type = self.__is_youtube_link(url)
        if not is_valid_link:
            error_console.print("❌ Invalid link")
            sys.exit(1)

        return is_valid_link, link_type.lower()

    def __is_youtube_link(self, link: str) -> tuple[bool, str]:
        """
        Check if the given link is a YouTube video, playlist, or shorts link.

        Args:
            link (str): The link to be checked.

        Returns:
            tuple[bool, str]: True if the link is a YouTube video, playlist, or shorts link,
            False otherwise. The second item of the tuple indicates the type of link found:
            'video', 'playlist', or 'shorts'.
        """
        is_video = self.__is_youtube_video(link)
        is_short = self.__is_youtube_shorts(link)
        is_playlist = self.__is_youtube_playlist(link)

        return (is_video, "video") if is_video \
            else (is_short, "short") if is_short \
            else (True, "playlist") if is_playlist\
            else (False, "unknown")

    def __is_youtube_shorts(self, link: str) -> bool:
        """
        Check if the given link is a YouTube shorts link.

        Args:
            link: The link to be checked.

        Returns:
            bool: True if the link is a YouTube shorts link, False otherwise.
        """
        # shorts_pattern = r"(?:https?:\/\/)?(?:www\.)?youtube\.com\/shorts\/([a-zA-Z0-9_-]+)"
        shorts_pattern = r"(?:https?:\/\/)?(?:www\.)?(?:youtube\.com\/(?:[^\/\n\s]+\/\S+\/|shorts\/|watch\?.*?v=))(?:(?:[^\/\n\s]+\/)?)([a-zA-Z0-9_-]+)"
        shorts_match = re.match(shorts_pattern, link)
        return bool(shorts_match)

    def __is_youtube_video(self, link: str) -> bool:
        """
        Check if the given link is a YouTube video.

        Args:
            link: The link to be checked.

        Returns:
            bool: True if the link is a YouTube video, False otherwise.
        """
        # video_pattern = re.compile(
        # r'(?:https?://)?(?:www\.)?(?:youtube\.com/(?:(?:live/?[a-zA-Z0-9_-]{11}\?si=)|(?:(?:watch\?v=)|(?:embed/))|youtu\.be/|youtube.com/share\?v=)([a-zA-Z0-9_-]{11}))')
        # video_pattern = re.compile(
        # r'(?:https?://)?(?:www\.)?(?:youtube\.com/(?:(?:watch\?v=)|(?:embed/))|youtu\.be/|youtube.com/share\?v=)([a-zA-Z0-9_-]{11})')
        # video_pattern = re.compile(
        #     # "https://www.youtube.com/watch?time_continue=1&v=dQw4w9WgXcQ"
        # r"^((?:https?:)?\/\/)?((?:www|m)\.)?((?:youtube\.com|youtu.be))(\/(?:[\w\-]+\?v=|embed\/|v\/)?)([\w\-]+)(\S+)?$"
        # )
        video_pattern = re.compile(
            r"^(?:https?://)?(?:www\.)?(?:youtube(?:-nocookie)?\.com/(?:(watch\?v=|watch\?feature\=share\&v=)|embed/|v/|live_stream\?channel=|live\/)|youtu\.be/)([a-zA-Z0-9_-]{11})"
        )

        return bool(video_pattern.match(link))

    def __is_youtube_playlist(self, link: str) -> bool:
        """
        Check if the given link is a YouTube playlist.

        Args:
            link: The link to be checked.

        Returns:
            bool: True if the link is a YouTube playlist, False otherwise.
        """
        playlist_pattern = r"^(?:https?:\/\/)?(?:www\.)?youtube\.com\/playlist\?list=([a-zA-Z0-9_-]+)"
        watch_with_playlist_pattern = r"^(?:https?:\/\/)?(?:www\.)?youtube\.com\/watch\?v=[a-zA-Z0-9_-]{11}&list=([a-zA-Z0-9_-]+)"

        return bool(re.match(playlist_pattern, link)) or bool(re.match(watch_with_playlist_pattern, link))

    def __is_youtube_video_id(self, video_id: str) -> bool:
        """
        Check if the given string is a valid YouTube video ID.

        Args:
            video_id: The string to be checked.

        Returns:
            bool: True if the string is a valid YouTube video ID, False otherwise.
        """
        return len(video_id) == 11 and all(
            c in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-" for c in video_id
        )



"""
This module contains the setup configuration for the pyutube package.
"""


from setuptools import find_packages, setup
from pyutube.utils import __version__


# Read the README file to use as the long description
with open("README.md", "r", encoding="utf-8") as f:
    description = f.read()

# Setup configuration
setup(
    name="pyutube",

    version=__version__,

    author="Ebraheem Alhetari",

    author_email="hetari4all@gmail.com",

    description="Awesome CLI to download YouTube videos (as video or audio)/shorts/playlists from the terminal",

    long_description=description,

    long_description_content_type="text/markdown",

    keywords=[
        "youtube",
        "download",
        "cli",
        "pyutube",
        "pytubefix",
        "pytube",
        "youtube-dl",
    ],

    license="MIT",

    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],


    include_package_data=True,

    python_requires=">=3.6",

    install_requires=[
        "pytubefix",
        "inquirer",
        "yaspin",
        "typer",
        "requests",
        "rich",
        "inquirer",
        "termcolor",
        "moviepy",
    ],


    entry_points={
        "console_scripts": [
            "pyutube=pyutube:cli.app",
        ],
    },


    project_urls={
        "Author": "https://github.com/Hetari",
        "Homepage": "https://github.com/Hetari/pyutube",
        "Bug Tracker": "https://github.com/Hetari/pyutube/issues",
        "Source Code": "https://github.com/Hetari/pyutube",
        "Documentation": "https://github.com/Hetari/pyutube",
    },

    platforms=["Linux", "Windows", "MacOS"],
    packages=find_packages()
)
